/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
//(C) PicoPeta Simputers Pvt. Ltd. (2003)
// PhotoAlbum

// Programmer: Santhosh N

#include "plib.h"

unsigned _rmask = 0xff, _gmask = 0xff, _bmask = 0xff;

void _get_visinfo(Display *disp)
{
	XVisualInfo *vi_in, vi_out;
     int nvis, i, j, n;
     unsigned depth, bpp;

	if(_rmask == 0xff && _gmask == 0xff && _bmask == 0xff) {
		vi_out.screen = DefaultScreen(disp);
	     vi_in = XGetVisualInfo(disp, VisualScreenMask, &vi_out, &nvis);

		if (vi_in) {
			for (i=0; i<nvis; i++) {
				XPixmapFormatValues *pf = XListPixmapFormats(disp, &n);
				depth = vi_in[i].depth;
				bpp = 0;
				for (j = 0; j < n; j++) {
					if (pf[j].depth == depth) {
						bpp = pf[j].bits_per_pixel;
						break;
					}
				}

				if (bpp == 16) {
					_rmask = vi_in[i].red_mask;
					_gmask = vi_in[i].green_mask;
					_bmask = vi_in[i].blue_mask;
					break;
				}
			}

			XFree(vi_in);
     	}
	}
}

void _pimg_imglst_add(struct Image **start, struct Image *node)
{
	struct Image *t;

	t = *start;

	if(t==NULL) {
		*start = node;
	} else {
		while(t->next != NULL)
			t = t->next;
		t->next = node;
	}
}

void _pimg_rawimglst_add(struct RawImage **start, struct RawImage *node)
{
	struct RawImage *t;

	t = *start;

	if(t==NULL) {
		*start = node;
	} else {
		while(t->next != NULL)
			t = t->next;
		t->next = node;
	}
}

static void convert_raw_to_x(Display *disp, int depth, unsigned char *from, unsigned char *to, int width, int height)
{
	unsigned x, y, i, t, llen;
	unsigned char r, g, b;
	unsigned long p1, p2;

	switch(depth) {
		case 4:
			for(y=0; y<height; y++) {
				for(x=0, i=0; x<width*3; i++, x+=6) {
					llen = y*width*3+x;
					p1 = from[llen]*11 + from[llen+1]*16 + from[llen+2]*5; 
					p2 = from[llen+3]*11 + from[llen+4]*16 + from[llen+5]*5; 

					p1 = (((p2>>9)<<4)|(p1>>9))&0xff;

					memcpy(to + y*width + i, &p1, 1);
				}
			}
			break;
		case 12:
			for(y=0; y<height; y++) {
				for(x=0, i=0; x<width*3; i+=2, x+=3) {
					llen = y*width*3+x;
					r = from[llen]; g = from[llen+1]; b = from[llen+2];

					t = ((r>>4)<<8) + ((g>>4)<<4) + (b>>4);

					memcpy(to + y*width*2 + i, &t, 2);
				}
			}
			break;
		case 16:
			if(_rmask==0x00f800 && _gmask==0x0007e0 && _bmask==0x00001f) {
				for(y=0; y<height; y++) {
					for(x=0, i=0; x<width*3; i+=2, x+=3) {
						llen = y*width*3+x;
						r = from[llen]; g = from[llen+1]; b = from[llen+2];
						t = ((r>>3)<<11) + ((g>>2)<<5) + (b>>3);
						memcpy(to + y*width*2 + i, &t, 2);
					}
				}
			} else {
				for(y=0; y<height; y++) {
					for(x=0, i=0; x<width*3; i+=2, x+=3) {
						llen = y*width*3+x;
						r = from[llen]; g = from[llen+1]; b = from[llen+2];
						t = ((r>>2)<<10) + ((g>>3)<<5) + (b>>3);
						memcpy(to + y*width*2 + i, &t, 2);
					}
				}
			}
			break;
	}
}

int pimg_raw_moveto_x(Display *disp, struct RawImage **rimg, struct Image **ximg)
{
	struct RawImage *t, *r;
	struct Image *img;
	int depth;
	unsigned char *tmp;

	depth = DefaultDepth(disp, DefaultScreen(disp));

	if(depth!=4 && depth!=12 && depth!=16)
		return EBADDEP;

	_get_visinfo(disp);

     t = *rimg;
	*rimg = NULL;

	while(t) {
		convert_raw_to_x(disp, depth, t->img, t->img, t->width, t->height);

		// Create the Image and add to the list
		img = (struct Image*)malloc(sizeof(struct Image));
		img->info = t->info;
		img->delay = t->delay;
		img->x = t->x;
		img->y = t->y;
		img->width = t->width;
		img->height = t->height;
		img->next = NULL;
		img->img = img->mimg = NULL;
		
		if(depth==4) {
			tmp = (unsigned char*)realloc(t->img, t->width*(t->height+1));
			if(tmp)
				t->img = tmp;
			img->img = XCreateImage(disp, CopyFromParent, depth, ZPixmap, 0,\
								t->img, img->width, img->height, 8, img->width);
		} else {
			tmp = (unsigned char*)realloc(t->img, t->width*(t->height+1)*2);
			if(tmp)
				t->img = tmp;
			img->img = XCreateImage(disp, CopyFromParent, depth, ZPixmap, 0,\
								t->img, img->width, img->height, 16, img->width*2);
		}
	
		if(t->mimg) {
		}

		_pimg_imglst_add(ximg, img);

		r = t;
		t = t->next;
		free(r);
	}

	return 0;
}

int pimg_raw_copyto_x(Display *disp, struct RawImage **rimg, struct Image **ximg)
{
	struct RawImage *t;
	struct Image *img;
	int depth;

	depth = DefaultDepth(disp, DefaultScreen(disp));

	if(depth!=4 && depth!=12 && depth!=16)
		return EBADDEP;

	_get_visinfo(disp);

     t = *rimg;

	while(t) {
		convert_raw_to_x(disp, depth, t->img, t->img, t->width, t->height);

		// Create the Image and add to the list
		img = (struct Image*)malloc(sizeof(struct Image));
		img->info = t->info;
		img->delay = t->delay;
		img->x = t->x;
		img->y = t->y;
		img->width = t->width;
		img->height = t->height;
		img->next = NULL;
		img->img = img->mimg = NULL;
		
		if(depth==4) {
			img->img = XCreateImage(disp, CopyFromParent, depth, ZPixmap, 0,\
								t->img, img->width, img->height, 8, img->width);
		} else {
			img->img = XCreateImage(disp, CopyFromParent, depth, ZPixmap, 0,\
								t->img, img->width, img->height, 16, img->width*2);
		}
	
		if(t->mimg) {
		}

		_pimg_imglst_add(ximg, img);

		t = t->next;
	}

	return 0;
}

int pimg_load(char *file, Display *disp, struct Image **ximg, unsigned short iattr)
{
	unsigned char buff[15];
	int ret=ENOTIMG;
	unsigned depth;
	FILE *fp;
	struct RawImage *rimg=NULL;

	depth = DefaultDepth(disp, DefaultScreen(disp));

	if(depth!=4 && depth!=12 && depth!=16)
		return EBADDEP;
			
	*ximg = NULL;

	fp = fopen(file, "rb");
	if(fp==NULL)
		return ENOFILE;

	_get_visinfo(disp);

	// Read the initial few bytes, to figure out the image format
	fread(buff, 1, 9, fp);
	rewind(fp);

	switch(buff[0]) {
		case 0xff:	// For JPEG
			if(buff[1]==0xd8)
				ret = _load_jpeg_rawimage(fp, &rimg);

			if(ret==0) {
				ret = FORMAT_JPEG;
				if(iattr&DITHER)
					pimg_raw_dither(depth, &rimg);
				pimg_raw_moveto_x(disp, &rimg, ximg);
			}

			break;
		case 0x89:	// For PNG
			if(buff[1]=='P' && buff[2]=='N' && buff[3]=='G')
				ret = _load_png_rawimage(fp, &rimg, iattr);
			
			if(ret==0) {
				ret = FORMAT_PNG;
				if(iattr&DITHER)
					pimg_raw_dither(depth, &rimg);
				pimg_raw_moveto_x(disp, &rimg, ximg);
			}
			
			break;
		case 'B':		// For BMP
			buff[2] = 0;
			if(!strcmp("BM",buff))
				ret = _load_bmp_rawimage(fp, &rimg);
			
			if(ret==0) {
				ret = FORMAT_BMP;
				if(iattr&DITHER)
					pimg_raw_dither(depth, &rimg);
				pimg_raw_moveto_x(disp, &rimg, ximg);
			}
			
			break;
		case 'G':		// For GIF
			buff[4] = 0;
			if(!strcmp("GIF8",buff))
				ret = load_gif_image(fp, disp, ximg, iattr);
			
			if(ret==0)
				ret = FORMAT_GIF;
			
			break;
		case '/':		// For XPM
			buff[9] = 0;
			if(!strcmp("/* XPM */",buff))
				ret = load_xpm_image(file, fp, disp, ximg, iattr);
			
			if(ret==0)
				ret = FORMAT_XPM;
			
			break;
		default:
			ret = ENOSUPP;
			break;
	}

	fclose(fp);

	return ret;
}

int pimg_save(char *filename, Display *disp, struct Image **img, int image_format, unsigned quality)
{
	_get_visinfo(disp);

	if(image_format==FORMAT_JPEG) {
		if(quality<1 || quality>100)
			return EBADVAL;
		return _save_jpeg_image(filename, disp, img, quality);
	}

	if(image_format==FORMAT_PNG) {
		return _save_png_image(filename, disp, img);
	}

	if(image_format==FORMAT_PNM) {
		return _save_pnm_image(filename, disp, img);
	}

	if(image_format==FORMAT_PGM) {
		return _save_pgm_image(filename, disp, img);
	}

	return ENOSUPP;
}

void pimg_free(struct Image **ximg)
{
	struct Image *t;

	while(*ximg!=NULL) {
		t = *ximg;
		*ximg = (*ximg)->next;
		XDestroyImage(t->img);
		if(t->mimg!=NULL)
			XDestroyImage(t->mimg);
		free(t);
	}
}

int pimg_raw_load(char *file, struct RawImage **rimg, unsigned short iattr)
{
	unsigned char buff[15];
	int ret=ENOTIMG;
	FILE *fp;

	*rimg = NULL;

	fp = fopen(file, "rb");
	if(fp==NULL)
		return ENOFILE;

	// Read the initial few bytes, to figure out the image format
	fread(buff, 1, 9, fp);
	rewind(fp);

	switch(buff[0]) {
		case 0xff:	// For JPEG
			if(buff[1]==0xd8)
				ret = _load_jpeg_rawimage(fp, rimg);

			if(ret==0)
				ret = FORMAT_JPEG;

			break;
		case 0x89:	// For PNG
			if(buff[1]=='P' && buff[2]=='N' && buff[3]=='G')
				ret = _load_png_rawimage(fp, rimg, iattr);
			
			if(ret==0)
				ret = FORMAT_PNG;
			
			break;
		case 'B':		// For BMP
			buff[2] = 0;
			if(!strcmp("BM",buff))
				ret = _load_bmp_rawimage(fp, rimg);
			
			if(ret==0)
				ret = FORMAT_BMP;
			
			break;
		default:
			ret = ENOSUPP;
			break;
	}

	fclose(fp);

	return ret;
}

void pimg_raw_free(struct RawImage **rimg)
{
	struct RawImage *t;

	while(*rimg!=NULL) {
		t = *rimg;
		*rimg = (*rimg)->next;
		if(t->img);
			free(t->img);
		if(t->mimg)
			free(t->mimg);
		free(t);
	}
}

char *pimg_errmsg(int err_val)
{
	switch(err_val) {
		case EBADIMG:
			return "Bad Image";
		case EBIGIMG:
			return "Big Image";
		case ENOTIMG:
			return "Not an Image";
		case ENOFILE:
			return "File does not exist";
		case ENOMEMR:
			return "No enough memory";
		case ENOSUPP:
			return "Image format not supported";
		case EBADDEP:
			return "Depth of the display should be 4/12/16 bits";
		default:
			return "Unknown error value";
	}
}
